/*
 *    gauss.C --- Contains the functions do detect the identity foam and
 *                subsequently applay Gauss' Lemma.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012/2013 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <list>
#include <vector>
#include <deque>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"

bool chainComplex::gauss(unsigned mapIdx, unsigned rowIdx, unsigned colIdx,
        std::deque<std::vector<unsigned> > *gaussStack, bool check) {
    two* theMap = maps.at(mapIdx);
    if (check) {
        if (!(theMap->rowsSM(rowIdx) && theMap->colsSM(colIdx))) return false;
        lcfoams* l = theMap->at(rowIdx, colIdx);
        if (! l) return false;
        if (! l->dualEqualsInverse(*spaces.at(mapIdx)->webs.at(colIdx)))
            return false;
    }

    ENTRYCONT<SparseEntry<lcfoams> > * LCol = theMap->colsSM(colIdx);
    ENTRYCONT<SparseEntry<lcfoams> > * LRow = theMap->rowsSM(rowIdx);
    foam* idFoam = theMap->at(rowIdx, colIdx)->firstFoam()->copy(true);
    idFoam->inverseCoefficient();

    for (ENTRYCONT<SparseEntry<lcfoams> >::iterator c = LCol->begin();
            c != LCol->end(); ++c) {
        if (c->idx == rowIdx) continue;
        for (ENTRYCONT<SparseEntry<lcfoams> >::iterator r = LRow->begin();
                r != LRow->end(); ++r) {
            if (r->idx == colIdx) continue;

            lcfoams *lNew = r->val->copy();
            lNew->glueLcfoams(idFoam->copy());
            lNew->glueLcfoams(c->val->copy());
            lNew->changeSign();
            bool pushItBack = false;

            if (! lNew->simplify()) {
                lcfoams* l = theMap->at(c->idx, r->idx);
                if (l) {
                    l->foams.splice(l->foams.end(), lNew->foams);
                    if (l->simplify()) {
                        theMap->erase(c->idx, r->idx);
                        delete l;
                    } else
                        pushItBack = l->dualEqualsInverse(
                                *spaces.at(mapIdx)->webs.at(r->idx));
                    delete lNew;
                }
                else {
                    theMap->set(c->idx, r->idx, lNew);
                    pushItBack = lNew->dualEqualsInverse(
                            *spaces.at(mapIdx)->webs.at(r->idx));
                }
            } else delete lNew;
            if (gaussStack && pushItBack) {
                std::vector<unsigned> v;
                v.push_back(mapIdx);
                v.push_back(c->idx);
                v.push_back(r->idx);
                gaussStack->push_back(v);
            }
        }
    }
    idFoam->deleteContents();
    delete idFoam;

    theMap->eraseRow(rowIdx, &lcfoams::kill);
    theMap->eraseCol(colIdx, &lcfoams::kill);
    spaces.at(mapIdx + 1)->eraseWeb(rowIdx);
    spaces.at(mapIdx)->eraseWeb(colIdx);
    if (mapIdx > 0) maps.at(mapIdx - 1)->eraseRow(colIdx, &lcfoams::kill);
    if (mapIdx < maps.size() - 1) maps.at(mapIdx + 1)->eraseCol(rowIdx,
            &lcfoams::kill);
    return true;
}

bool facet::mayBePartOfIdentity() const {
    unsigned const bsize = boundaries.size();
    if ((dots != 0) || (bsize == 0) || (bsize > 2)) return false;
    unsigned oPointsContained = 0, iEdgesContained = 0;
    bool touchedBelow = false, touchedAbove = false;
    for (BOUNDARIESCONT<EDGECONT<edge*> >::const_iterator clleit =
            boundaries.begin(); clleit != boundaries.end(); ++clleit)
        for (EDGECONT<edge*>::const_iterator cleit = clleit->begin();
                cleit != clleit->end(); ++cleit) {
            if ((*cleit)->isInner) ++iEdgesContained;
            else {
                oEdge * oe = (*cleit)->asOEdge();
                if (oe->isBelow) touchedBelow = true;
                else touchedAbove = true;
                if (! oe->isCircle())
                    for (unsigned k = 0; k < 2; ++k)
                        if (! oe->points[k]->isInner) ++oPointsContained;
            }
        }
    return ( ( touchedBelow && touchedAbove) && 
         (((bsize == 2) && (oPointsContained == 0) && (iEdgesContained == 0)) ||
           ((bsize == 1) && (oPointsContained / 2 + iEdgesContained == 2))) );
}

bool foam::dualEqualsInverse() const {
    if (! coefficient.isInvertible()) return false;
    for (FACETCONT<facet*>::const_iterator cfit = facets.begin();
            cfit != facets.end(); ++cfit)
        if (! (*cfit)->mayBePartOfIdentity() ) return false;
    return true;
}

bool lcfoams::dualEqualsInverse(web &below) const {
    below.setEdgesToBelow();
    const bool result = ((foams.size() == 1) &&
            (firstFoam()->dualEqualsInverse()));
    below.setEdgesToAbove();
    return result;
}
