/*
 *    geometry.C --- Contains those functions which are particular to the
 *                   geometry of webs and foams.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012/2013 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <list>
#include <vector>
#include <deque>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"

void lcfoams::glueLcfoams(lcfoams* other, bool dualizeIt, bool deleteIt) {
    FACETCONT<foam*> oldFoams;
    oldFoams.splice(oldFoams.begin(), foams);
    for (FACETCONT<foam*>::iterator it = oldFoams.begin();
            it != oldFoams.end(); ++it) {
        for (FACETCONT<foam*>::iterator itOther = other->foams.begin();
                itOther != other->foams.end(); ++itOther) {
            foam* f = (*it)->copy();
            f->glueFoam((*itOther)->copy(dualizeIt));
            foams.push_back(f);
        }
        (*it)->deleteContents();
        delete *it;
    }
    if (deleteIt) {
        other->deleteContents();
        delete other;
    }
}

void chainComplex::composeEverything(web* W1, web* W2, foam* map) {
    std::vector<std::vector<foam*> > fade;
    std::vector<std::vector<web*> > smallWebs;
    fade.resize(spaces.size());
    smallWebs.resize(spaces.size());

    for (unsigned int spacesI = 0, lastBit = 0; spacesI < spaces.size();
            ++spacesI) {
        one* thisOne = spaces.at(spacesI);
        const unsigned int max = thisOne->webs.size();
        const unsigned int realMax = thisOne->realSize();
        fade.at(spacesI).resize(max, 0);
        smallWebs.at(spacesI).resize(max, 0);
        for (unsigned int WebI = 0; WebI < max; ++WebI) {
            if (thisOne->webs.at(WebI)) {
                web* Wx = (WebI < max - lastBit) ? W1 : W2;
                foam* id = new foam;
                smallWebs.at(spacesI).at(WebI) = Wx->copy(id);
                fade.at(spacesI).at(WebI) = id;
            }
        }
        lastBit = realMax - lastBit;
    }
    
    for (unsigned int mapsI = 0, lastBit = spaces.at(0)->webs.size();
            mapsI < maps.size(); ++mapsI) {
        MatrixIterator<lcfoams> it = maps.at(mapsI)->begin();
        while (! it.isFinished()) {
            foam* toGlue = fade.at(mapsI).at(it.c())->copy(true);
            if (map && it->doGlueSS) {
                toGlue->glueFoam(map->copy());
                it->doGlueSS = false;
            }
            toGlue->glueFoam(fade.at(mapsI + 1).at(it.r())->copy());
            it->composeLcFoams(toGlue, * spaces.at(mapsI)->webs.at(it.c()),
                    * smallWebs.at(mapsI).at(it.c()));
            if (! maps.at(mapsI)->lcfSimplify(&it)) it.step();
        }
        lastBit = spaces.at(mapsI + 1)->webs.size() - lastBit;
    }

    for (std::vector<std::vector<foam*> >::iterator it = fade.begin();
            it != fade.end(); ++it)
        for (std::vector<foam*>::iterator it2 = it->begin();
                it2 != it->end(); ++it2) {
            if (!*it2) continue;
            (*it2)->deleteContents();
            delete *it2;
        }
    for (unsigned int spacesI = 0; spacesI != spaces.size(); ++spacesI) {
        one* thisOne = spaces.at(spacesI);
        for (unsigned int k = 0; k < thisOne->webs.size(); ++k)
            if (thisOne->webs.at(k))
                thisOne->webs.at(k)->composeWeb(smallWebs.at(spacesI).at(k));
    }
}

lcfoams* lcfoams::copy(bool dualizeIt) const {
    lcfoams* result = new lcfoams();
    for (FACETCONT<foam*>::const_iterator i = foams.begin();
            i != foams.end(); ++i)
        result->foams.push_back((*i)->copy(dualizeIt));
    return result;
}

web* web::copy(foam* id) {
    if (id) id->coefficient = ratio(1);
    web *neww = new web();
    neww->qShift = qShift;

    for (OEDGECONT<oEdge*>::iterator i = edges.begin(); i != edges.end(); ++i) {
        oEdge *newe = new oEdge();
        (*i)->copyTag = newe;
        for (int j = 0; j < 2; ++j)
            newe->points[j] = (*i)->points[j];
        neww->edges.push_back(newe);
        if (id) (*i)->idFacet = 0;
    }

    if (id)
        for (IPOINTCONT<iPoint*>::iterator i = points.begin();
                i != points.end(); ++i)
            (*i)->idSingular = 0;

    for (IPOINTCONT<iPoint*>::iterator i = points.begin();
            i != points.end(); ++i) {
        (*i)->copyTag = new iPoint();
        for (int j = 0; j < 3; ++j)
            (*i)->copyTag->edges[j] = (*i)->edges[j];
        neww->points.push_back((*i)->copyTag);
        if (id && !((*i)->isTransit())) {
            id->singulars.resize(id->singulars.size() + 1);
            iEdge* e = new iEdge(); 
            *(id->singulars.rbegin()) = e;
            if ((*i)->isIncoming()) {
                e->points[0] = *i;
                e->points[1] = (*i)->copyTag;
            } else {
                e->points[1] = *i;
                e->points[0] = (*i)->copyTag;
            }
            (*i)->idSingular = e;
        }
    }

    if (id) {
        for (OEDGECONT<oEdge*>::iterator i = edges.begin();
                i != edges.end(); ++i)
            if (!((*i)->idFacet)) {
                if ((*i)->isCircle()) {
                    facet *f = new facet();
                    (*i)->idFacet = f;
                    id->facets.push_back(f);

                    f->dots = 0;
                    f->boundaries.resize(2);
                    f->boundaries.begin()->push_back(*i);
                    f->boundaries.rbegin()->push_back((*i)->copyTag);
                } else if ( (! (*i)->points[0]->isInner) ||
                        (! (*i)->pointAsI(0)->isTransit()) ) {
                    facet *f = new facet();
                    id->facets.push_back(f);

                    f->dots = 0;
                    f->boundaries.resize(1);
                    EDGECONT<edge*>* l = &*(f->boundaries.begin());
                    if ((*i)->points[0]->isInner)
                        l->push_back((*i)->pointAsI(0)->idSingular);
                    oEdge* j = *i;
                    while (j->points[1]->isInner &&
                            j->pointAsI(1)->isTransit()) {
                        l->push_back(j);
                        l->push_front(j->copyTag);
                        j->idFacet = f;
                        j = j->pointAsI(1)->otherEdge(*j);
                    }
                    j->idFacet = f;
                    l->push_back(j);
                    l->push_front(j->copyTag);
                    if (j->points[1]->isInner)
                        l->push_back(j->pointAsI(1)->idSingular);
                }
            }

        /* Work on circles made of several segments */
        for (OEDGECONT<oEdge*>::iterator i = edges.begin();
                i != edges.end(); ++i)
            if (!((*i)->idFacet)) {
                    facet *f = new facet();
                    id->facets.push_back(f);

                    f->dots = 0;
                    f->boundaries.resize(2);
                    EDGECONT<edge*>* l1 = &*(f->boundaries.begin());
                    EDGECONT<edge*>* l2 = &*(f->boundaries.rbegin());
                    oEdge* j = (*i)->pointAsI(1)->otherEdge(**i);
                    while (j != *i) {
                        l1->push_back(j);
                        l2->push_front(j->copyTag);
                        j->idFacet = f;
                        j = j->pointAsI(1)->otherEdge(*j);
                    }
                    j->idFacet = f;
                    l1->push_back(j);
                    l2->push_front(j->copyTag);
            }

        for (IPOINTCONT<iPoint*>::iterator i = points.begin();
                i != points.end(); ++i)
            if (! (*i)->isTransit())
                for (int j = 0; j < 3; ++j)
                    (*i)->idSingular->facets[j] = (*i)->edges[j]->idFacet;
    }

    for (OEDGECONT<oEdge*>::iterator i = neww->edges.begin();
            i != neww->edges.end(); ++i)
        if (! (*i)->isCircle())
            for (int j = 0; j < 2; ++j)
                if ((*i)->points[j]->isInner)
                    (*i)->points[j] = ((iPoint *)((*i)->points[j]))->copyTag;

    for (IPOINTCONT<iPoint*>::iterator i = neww->points.begin();
            i != neww->points.end(); ++i)
        for (int j = 0; j < 3; ++j)
            if ((*i)->edges[j])
                (*i)->edges[j] = (oEdge *)((*i)->edges[j]->copyTag);

    for (OEDGECONT<oEdge*>::iterator i = edges.begin(); i != edges.end(); ++i)
        (*i)->copyTag = 0;

    for (IPOINTCONT<iPoint*>::iterator i = points.begin();
            i != points.end(); ++i)
        (*i)->copyTag = 0;

    if (id->degree() != 0) {
        std::cerr << "\ndegree error while copying web.\n";
        throw 14;
    }
    lcfoams *l = new lcfoams(id);
    if (! l->dualEqualsInverse(*this)) {
        std::cerr << "A foam that should be identity is not.\n";
        throw 15;
    }
    delete l;
    return neww;
}

void iPoint::findGoals(iPoint* goal[], std::vector<oEdge*> w[]) const {
    const int idx = isIncoming() ? 0 : 1;
    for (int j = 0; j < 3; ++j) {
        oEdge* e = edges[j];
        if (e->points[idx]->isInner) {
            w[j].push_back(e);
            goal[j] = (iPoint*)(e->points[idx]);
            while (goal[j]->isTransit()) {
                e = goal[j]->otherEdge(*e);
                w[j].push_back(e);
                if (e->points[idx]->isInner)
                    goal[j] = (iPoint*)(e->points[idx]);
                else {
                    goal[j] = 0;
                    break;
                }
            }
            if (goal[j] == 0) continue;
        }
        else {
            goal[j] = 0;
            continue;
        }
    }
}

void web::composeWeb(web *other) {
    if (this == other) throw 24;
    qShift += other->qShift;
    edges.splice(edges.end(), other->edges);
    points.splice(points.end(), other->points);
    delete other;
    for (OEDGECONT<oEdge*>::iterator ioe = edges.begin();
            ioe != edges.end(); ++ioe)
        for (int i = 0; i < 2; ++i)
            if (! ((*ioe)->points[i]->isInner))
                for (OEDGECONT<oEdge*>::iterator ioe2 = (++ioe)--;
                        ioe2 != edges.end(); ++ioe2)
                    if ((*ioe2)->points[1 - i] == (*ioe)->points[i]) {
                        iPoint *newIPoint = new iPoint();
                        points.push_back(newIPoint);
                        (*ioe2)->points[1 - i] = (*ioe)->points[i] = newIPoint;
                        newIPoint->edges[i] = *ioe;
                        newIPoint->edges[1 - i] = *ioe2;
                        break;
                    }
}

void foam::composeFoam(foam *other, const web & below, const web & below2) {
    int degShouldBe = degree() + other->degree();

    if (this == other) throw 25;

    singulars.splice(singulars.end(), other->singulars);
    facets.splice(facets.end(), other->facets);
    coefficient *= other->coefficient;
    delete other;

    for (FACETCONT<facet*>::iterator i = facets.begin();
            i != facets.end(); ++i) {
        (*i)->heir = 0;
        for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
                (*i)->boundaries.begin(); ile != (*i)->boundaries.end(); ++ile)
            for (EDGECONT<edge*>::iterator ie = ile->begin();
                    ie != ile->end(); ++ie)
                if ( ! (*ie)->isInner ) {
                    (*ie)->asOEdge()->isBelow = false;
                    for (int j = 0; j < 2; ++j)
                        if ( ! (*ie)->asOEdge()->points[j]->isInner )
                            (*ie)->asOEdge()->points[j]->asOPoint()->isGlued =
                                false;
                }
    }

    below.setEdgesToBelow();
    below2.setEdgesToBelow();

    FACETCONT<facet*> facetsToErase;

    for (FACETCONT<facet*>::iterator i = facets.begin(); i != facets.end(); ++i)
        for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
                (*i)->boundaries.begin();
                ile != (*i)->boundaries.end(); ++ile) {
            bool glued = false;
            for (EDGECONT<edge*>::iterator ie = ile->begin();
                    ie != ile->end(); ++ie) {
                /* for the case that you have to glue twice in a row along
                 * an endpoint of an edge
                 */
                if (glued) --ie; 
                glued = false;
                if ((*ie)->isInner) continue;

                oEdge *thisOE = (*ie)->asOEdge();

                for (int j = 0; j < 2; ++j) {
                    thisOE = (*ie)->asOEdge();
                    bool isBelow = thisOE->isBelow;
                    bool hasBeenReversed = false;
                    if (thisOE->points[j]->isInner) continue;
                    oPoint *o = thisOE->points[j]->asOPoint();
                    if (o->isGlued) continue;
                    if (isBelow xor (j == 1)) {
                        if (ie != ile->begin()) {
                            /* this means there is nothing to glue to,
                             * otherwise that edge would have already been found
                             */
                            continue;
                        }
                        ile->splice(ile->begin(), *ile, --ile->end());
                        ie = ile->begin();
                        isBelow = !isBelow;
                        hasBeenReversed = true;
                    }
                    EDGECONT<edge*>::iterator insBefore = ie;
                    ++insBefore;
                    for (FACETCONT<facet*>::iterator i2 = i;
                            i2 != facets.end(); ++i2) {
                        for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile2 =
                                (i2 == i) ? ile : (*i2)->boundaries.begin();
                                ile2 != (*i2)->boundaries.end(); ++ile2) {
                            bool foundFirst = false;
                            for (EDGECONT<edge*>::iterator ie2 =
                                    (ile2 == ile) ? (++ie)-- : ile2->begin();
                                    ie2 != ile2->end(); ++ie2) {
                                if ((*ie2)->isInner) continue;
                                oEdge *thisOE2 = (*ie2)->asOEdge();

                                const bool isBelow2 = thisOE2->isBelow;
                                if (isBelow != isBelow2) {
                                    foundFirst = true;
                                    continue;
                                }
                                if (thisOE2->points[1 - j] == o) {
                                    glued = true;
                                    degShouldBe += 2;
                                    o->isGlued = true;
                                    if (! foundFirst) {
                                        if (ie2 != ile2->begin())
                                            throw 26;
                                        ile2->splice(ile2->begin(), *ile2,
                                                --ile2->end());
                                        ie2 = ++ile2->begin();
                                    }
                                    if (&*ile2 == &*ile) {

                                        (*i)->boundaries.resize(
                                                (*i)->boundaries.size() + 1);
                                        EDGECONT<edge*> *le =
                                            &*((*i)->boundaries.rbegin());
                                        le->splice(le->begin(), *ile,
                                                insBefore, ie2);
                                    } else {
                                        ile->splice(insBefore, *ile2, ie2,
                                                ile2->end());
                                        ile->splice(insBefore, *ile2);
                                        (*i2)->boundaries.erase(ile2);
                                        if (*i == *i2) {
                                            coefficient *= ratio(-3);
                                            (*i)->dots += 2;
                                        } else {
                                            (*i)->dots += (*i2)->dots;
                                            (*i)->boundaries.splice(
                                                    (*i)->boundaries.end(),
                                                    (*i2)->boundaries);
                                            (*i2)->heir = *i;
                                            facetsToErase.splice(
                                                    facetsToErase.end(),
                                                    facets, i2);
                                        }
                                    }
                                    break;
                                }
                            }
                            if (glued) break;
                        }
                        if (glued) break;
                    }
                    if ((!glued) && hasBeenReversed) {
                        ile->splice(ile->end(), *ile, ile->begin());
                        ie = ile->begin();
                    }
                }
            }
        }

    for (SINGULARCONT<iEdge*>::iterator iie = singulars.begin();
            iie != singulars.end(); ++iie)
        for (int j = 0; j < 3; ++j) {
            while ((*iie)->facets[j]->heir) {
                (*iie)->facets[j] = (*iie)->facets[j]->heir;
            }
        }

    for (FACETCONT<facet*>::iterator fit = facetsToErase.begin();
            fit != facetsToErase.end(); ++fit)
        delete *fit;

    below.setEdgesToAbove();
    below2.setEdgesToAbove();

    if (degShouldBe != degree()) {
        std::cerr << "Degree error while composing: degree should be " <<
            degShouldBe << ", but is " << degree() << "\n";
        throw 28000;
    }
}

void foam::glueFoam(foam *other) {
    /* 1. set heirs of iEdges
     * 2. make heirs "transitive"
     * 3. update iEdge info in facets
     * 4. glue facets along every edge
     */
    const int degShouldBe = degree() + other->degree();
    if (this == other) throw 27;
    coefficient *= other->coefficient;
    SINGULARCONT<iEdge*> toDelete;
    singulars.splice(singulars.begin(), other->singulars);
    SINGULARCONT<iEdge*>::iterator iie = singulars.begin();
    while (iie != singulars.end()) {
        bool glued = false;
        if (!((*iie)->isCircle())) {
            bool breakOut = false;
            for (int j = 0; j < 2; ++j)
                if (!breakOut) for (SINGULARCONT<iEdge*>::iterator kie = iie;
                        kie != singulars.end(); ++kie)
                    if ( (!((*kie)->isCircle())) && ((*kie)->points[1 - j] ==
                                (*iie)->points[j]) ) {
                        if (kie == iie) {
                            (*kie)->points[0] = (*kie)->points[1] = 0;
                            breakOut = true;
                            glued = true;
                            break;
                        } else {
                            (*iie)->points[j] = (*kie)->points[j];
                            (*kie)->toBeDeleted = true;
                            toDelete.splice(toDelete.end(), singulars, kie);
                            glued = true;
                            break;
                        }
                    }
        }
        if (! glued) ++iie;
    }

    facets.splice(facets.begin(), other->facets);
    delete other;

    FACETCONT<facet*> facetsToErase;

    for (FACETCONT<facet*>::iterator i = facets.begin();
            i != facets.end(); ++i) {
        BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
            (*i)->boundaries.begin();
        while (ile != (*i)->boundaries.end()) {
            bool deletePrev;
            EDGECONT<edge*>::iterator ie = ile->begin();
            while (ie != ile->end()) {
                deletePrev = false;
                if ((*ie)->isInner) {
                    ++ie;
                    continue;
                }
                bool isGlued = false;
                for (FACETCONT<facet*>::iterator i2 = i; i2 != facets.end();
                        ++i2) {
                    for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile2 =
                            ( (i2 == i) ? ile : (*i2)->boundaries.begin() );
                            ile2 != (*i2)->boundaries.end(); ++ile2) {

                        for (EDGECONT<edge*>::iterator ie2 =
                                ( (ile2 == ile) ? (++ie)-- : ile2->begin() );
                                ie2 != ile2->end(); ++ie2) {
                            if (*ie == *ie2) {
                                isGlued = true;
                                if (ile == ile2) {
                                    /* same boundary glue --> split boundary */
                                    (*i)->boundaries.resize(
                                            (*i)->boundaries.size() + 1);
                                    EDGECONT<edge*> *l =
                                        &*((*i)->boundaries.rbegin());
                                    l->splice(l->begin(), *ile, ie, ie2);
                                    ie = ie2;
                                    l->pop_front();
                                    deletePrev = true;
                                } else { /* merge */
                                    if (*i != *i2) {
                                        (*i)->dots += (*i2)->dots;
                                        (*i)->boundaries.splice(
                                                (*i)->boundaries.end(),
                                                (*i2)->boundaries);
                                        (*i2)->heir = *i;
                                        facetsToErase.splice(
                                                facetsToErase.end(),
                                                facets, i2);
                                    } else {
                                        /* same facet
                                         * different boundary glue --> genus
                                         */
                                        (*i2)->dots += 2;
                                        coefficient *= ratio(-3);
                                    }
                                    ile->splice(ie, *ile2, ie2, ile2->end());
                                    ile->splice(ie, *ile2);
                                    ile->erase(ie);
                                    ie = ie2;
                                    deletePrev = true;
                                    (*i2)->boundaries.erase(ile2);
                                }
                            }
                            if (isGlued) break;
                        }
                        if (isGlued) break;
                    }
                    if (isGlued) break;
                }
                if (deletePrev) ie = ile->erase(ie);
                else ++ie;
            }
            if (ile->empty()) ile = (*i)->boundaries.erase(ile);
            else ++ile;
        }
    }

    for (FACETCONT<facet*>::iterator i = facets.begin(); i != facets.end(); ++i)
        for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator ile =
                (*i)->boundaries.begin();
                ile != (*i)->boundaries.end(); ++ile) {
            EDGECONT<edge*>::iterator ie = ile->begin();
            while (ie != ile->end()) {
                EDGECONT<edge*>::iterator toErase = ie;
                if ((*ie)->toBeDeleted) {
                    ++ie;
                    ile->erase(toErase);
                } else ++ie;
            }
        }

    for (SINGULARCONT<iEdge*>::iterator iiie = singulars.begin();
            iiie != singulars.end(); ++iiie)
        for (int j = 0; j < 3; ++j)
            while ((*iiie)->facets[j]->heir) (*iiie)->facets[j] =
                (*iiie)->facets[j]->heir;

    if (degShouldBe != degree()) {
        std::cerr << "Degree error while gluing: degree should be " <<
            degShouldBe << ", but is " << degree() << "\n";
        throw 28;
    }

    for (SINGULARCONT<iEdge*>::iterator i = toDelete.begin();
            i != toDelete.end(); ++i)
        delete *i;
    for (FACETCONT<facet*>::iterator i = facetsToErase.begin();
            i != facetsToErase.end(); ++i)
        delete *i;
}

int foam::degree() { 
    int result = 0;

    for (FACETCONT<facet*>::iterator i = facets.begin();
            i != facets.end(); ++i) {
        result += 2 * (*i)->dots - 4 + 2 * (*i)->boundaries.size();
        for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator j =
                (*i)->boundaries.begin(); j != (*i)->boundaries.end(); ++j)
            for (EDGECONT<edge*>::iterator k = j->begin(); k != j->end(); ++k)
                if (!((*k)->isInner)) {
                    oEdge * e = (oEdge *)(*k);
                    if (!e->isCircle() && ((!e->points[0]->isInner) ||
                                (! e->pointAsI(0)->isTransit()))) {
                        result += 1;
                    }
                }
    }

    return result;
}

foam* foam::dualize() {
    for (SINGULARCONT<iEdge*>::iterator i = singulars.begin();
            i != singulars.end(); ++i) {
        iPoint *tmp = (*i)->points[0];
        (*i)->points[0] = (*i)->points[1];
        (*i)->points[1] = tmp;

        facet *ftmp = (*i)->facets[0];
        (*i)->facets[0] = (*i)->facets[1];
        (*i)->facets[1] = ftmp;
    }
    for (FACETCONT<facet*>::iterator j = facets.begin(); j != facets.end(); ++j)
        for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator i =
                (*j)->boundaries.begin(); i != (*j)->boundaries.end(); ++i)
            i->reverse();
    return this;
}

foam* foam::copy(bool dualizeIt) {
/* 1. Reset the copyTags
 * 2. Naive copying while setting copyTags
 * 3. Correcting the pointers using the copyTags
 */
    for (SINGULARCONT<iEdge*>::iterator i = singulars.begin();
            i != singulars.end(); ++i)
        (*i)->copyTag = 0;
    for (FACETCONT<facet*>::iterator i = facets.begin(); i != facets.end(); ++i)
        (*i)->copyTag = 0;

    foam* newf = new foam();
    newf->coefficient = coefficient;
    for (SINGULARCONT<iEdge*>::iterator i = singulars.begin();
            i != singulars.end(); ++i) {
        iEdge *newe = new iEdge(**i); /* shallow copy */
        if (dualizeIt) for (unsigned k = 0; k < 2; ++k)
            newe->points[k] = (*i)->points[1 - k];
        newf->singulars.push_back(newe);
        (*i)->copyTag = newe;
    }
    for (FACETCONT<facet*>::iterator i = facets.begin();
            i != facets.end(); ++i) {
        (*i)->copyTag = new facet(**i); /* shallow copy */
        if (dualizeIt)
            for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator j =
                    (*i)->copyTag->boundaries.begin();
                    j != (*i)->copyTag->boundaries.end(); ++j)
                j->reverse();
        newf->facets.push_back((*i)->copyTag);
    }

    for (FACETCONT<facet*>::iterator i = facets.begin(); i != facets.end(); ++i)
        if  (! (*i)->copyTag) throw 29;

    for (SINGULARCONT<iEdge*>::iterator i = newf->singulars.begin();
            i != newf->singulars.end(); ++i)
        for (int j = 0; j < 3; ++j) {
            (*i)->facets[j] = (*i)->facets[j]->copyTag;
        }
    for (FACETCONT<facet*>::iterator i = newf->facets.begin();
            i != newf->facets.end(); ++i)
        for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator j =
                (*i)->boundaries.begin(); j != (*i)->boundaries.end(); ++j)
            for (EDGECONT<edge*>::iterator k = j->begin(); k != j->end(); ++k)
                if ((*k)->copyTag) *k = (*k)->copyTag;

    if (degree() != newf->degree()) {
        std::cerr << "degree error while copying foam.\n";
        throw 30;
    }
    return newf;
}
