/*
 *    output.C --- Wraps some of the output.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012/2013 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <vector>
#include <deque>
#include <list>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"
#include "output.H"

unsigned printHomologyModN(unsigned int ccSpacesSize, int tShift, unsigned QIdx,
        int shiftedMinQ, std::vector<std::vector<unsigned int> > &homDims);

bool isChiral(unsigned int ccSpacesSize, int tShift, unsigned QIdx,
        int shiftedMinQ, const std::vector<std::vector<unsigned int> > &homDims,
        bool torsion);

void printHomology(unsigned int ccSpacesSize, int tShift, unsigned int QIdx,
        int shiftedMinQ, std::vector<std::vector<unsigned int> > &homDims,
        std::vector<torstruct> torsion, bool rational, bool reduced,
        bool addInfo) {
    std::cout << "Rational " << (reduced ? "reduced " : "") << "homology: ";
    unsigned totalDim = 
        printHomologyModN(ccSpacesSize, tShift, QIdx, shiftedMinQ, homDims);
    if (addInfo) {
        std::cout << "Total rank: " << totalDim << "\n";
        std::cout << "Rational homology is ";
        if (isChiral(ccSpacesSize, tShift, QIdx, shiftedMinQ, homDims, false))
            std::cout << "not self-dual => the link is chiral.\n";
        else
            std::cout << "self-dual.\n";
    }
    if (! rational) {
        if (torsion.empty())
            std::cout << "The homology is torsion-free.\n";
        for (std::vector<torstruct>::iterator vtit = torsion.begin();
                vtit != torsion.end(); ++vtit) {
            std::cout << vtit->mod << "-torsion: ";
            printHomologyModN(ccSpacesSize, tShift, QIdx,
                    shiftedMinQ, vtit->dim);
            if (addInfo) {
                std::cout << "The " << vtit->mod << "-torsion part of homology"
                    " is ";
                if (isChiral(ccSpacesSize, tShift, QIdx, shiftedMinQ,
                            vtit->dim, true))
                    std::cout << "not self-dual => the link is chiral.\n";
                else
                    std::cout << "self-dual.\n";
            }
        }
    }
}

bool isChiral(unsigned int ccSpacesSize, int tShift, unsigned QIdx,
        int shiftedMinQ, const std::vector<std::vector<unsigned int> > &homDims,
        bool torsion) {
    for (unsigned i = 0; i < homDims.size(); ++i)
        for (unsigned j = 0; j < homDims.at(i).size(); ++j)
            if (homDims.at(i).at(j)) {
                int otherI = (-i - 2 * tShift) + (torsion ? 1 : 0);
                int otherJ = (-j - shiftedMinQ);
                if ((otherI < 0) || (otherI >= (int)ccSpacesSize)
                        || (otherJ < 0) || (otherJ >= (int)QIdx)) return true;
                if (homDims.at(i).at(j) != homDims.at(otherI).at(otherJ))
                    return true;
            }
    return false;
}

unsigned printHomologyModN(unsigned int ccSpacesSize, int tShift, unsigned QIdx,
        int shiftedMinQ, std::vector<std::vector<unsigned int> > &homDims) {
    unsigned result = 0;
    bool isFirst = true;
    for (unsigned int i = 0; i < ccSpacesSize; ++i) {
        bool isFirst2 = true;
        for (unsigned int j = 0; j < QIdx; ++j) {
            const unsigned int dim = homDims.at(i).at(j);
            if (dim != 0) {
                result += dim;
                if (isFirst2) {
                    isFirst2 = false;
                    if (isFirst) isFirst = false;
                    else std::cout << " + ";
                    int degToPrint = (int)(i) + tShift;
                    switch (degToPrint)  {
                        case 0:  break;
                        case 1:  std::cout << "t"; break;
                        default:  std::cout << "t^" << degToPrint;
                    }
                    std::cout << "(";
                } else std::cout << " + ";
                if (dim != 1) std::cout << dim;
                int degToPrint = ((int)(2 * j) + shiftedMinQ);
                switch (degToPrint) {
                    case 0:  if (dim == 1) std::cout << "1"; break;
                    case 1:  std::cout << "q"; break;
                    default:  std::cout << "q^" << degToPrint;
                }
            }
        }
        if (! isFirst2) std::cout << ")";
    }
    std::cout << "\n";
    return result;
}

void printRasmussen(const std::vector<unsigned> &s, unsigned page2RasCandidate,
        int shift) {
    if (s.empty()) {
        std::cout << "Cannot get the s3-concordance invariant of"
            " multi-component links.\n";
    } else if (s.size() == 1)
        std::cout << "s_3-concordance invariant: " <<
            (int)(s.at(0)) + shift << "\n";
    else {
        std::cout << "There are several possibilites for the"
            " s_3-concordance invariant: ";
        for (unsigned i = 0; i < s.size(); ++i) {
            std::cout << (int)(s.at(i)) + shift;
            if (i == s.size() - 2)
                std::cout << " or ";
            else if (i < s.size() - 1)
                std::cout << ", ";
        }
        std::cout << "; if the spectral sequence to the filtered homology"
            " converges on the second page (as conjectured), the"
            " s_3-concordance is " << (int)page2RasCandidate + shift << ".\n";
    }
}

void printVVi(VECVECINT &v) {
    for (VECVECINT::iterator vviIt = v.begin(); vviIt != v.end(); ++vviIt) {
        if (vviIt != v.begin()) std::cout << ",";
        std::cout << "[";
        for (VECINT::iterator viIt = vviIt->begin();
                viIt + 1 != vviIt->end(); ++viIt) {
            if (viIt != vviIt->begin()) std::cout << ",";
            std::cout << *viIt + 1;
        }
        std::cout << "]";
    }
    std::cout << "].\n";
}

void printTitle()  {
    std::cout << "foamho, a sl3-homology calculator, version " << VERSION <<
        ". Have fun. Call with \"-h\" for help.\n";
}

void printHelp() {
    std::cout << "foamho, a sl3-homology calculator, version " << VERSION
        << ".\n\n" << "Usage: foamho [OPTIONS] braid | pd | dt NOTATION\n\n"
        << "For example, the following three commands all compute the\n"
        << "figure-8-knot's integral homology:\n\n"
        << "                    foamho braid aBaB\n"
        << "                    foamho pd [[4,2,5,1],[8,6,1,5],[6,3,7,4],"
        "[2,7,3,8]]\n"
        << "                    foamho dt [4,6,8,2]\n\n"
        << "Options:\n"
        << "-q                  Compute rational homology instead of "
                                                                   "integral.\n"
        << "-r                  Compute reduced homology instead of unreduced. "
                                                                "You may give\n"
        << "                    a number right after -r to indicate the marked"
                                                                     " strand\n"
        << "                    (useful for links).\n\n"
        << "-g                  Do not attempt to optimise the girth.\n"
        << "-v                  Display some progress information.\n"
        << "-vv                 Display more detailed progress information.\n"
        << "-t                  Display time and memory consumption.\n"
        << "-h                  Display this help message and exit.\n\n"
        << "Written in 2012/2013 by Lukas Lewark, lewark@math.jussieu.fr. "
                                                   "All feedback is welcome.\n";
}
