/*
 *    parser.C --- Parse the command line parameters and the
 *                 link diagram.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012/2013 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <vector>
#include <string.h>
#include <stdlib.h>

#include "macros.H"
#include "optionArray.H"
#include "dtToPd.H"
#include "parser.H"

int parseCommandLine(int argc, char* argv[], optionArray &oa) {
    oa.verbosityLevel = 0;
    oa.displayTimeMemory = false;
    oa.reduced = false;
    oa.rational = false;
    oa.optimiseGirth = true;
    oa.redStrand = 0;
    oa.addInfo = true;

    if (argc == 1) return OPTION_TITLE;
    if ((argc == 2) && (! strcmp(argv[1], "-h"))) return OPTION_HELP;

    for (int i = 1; i < argc - 2; ++i) {
        if (! strcmp(argv[i], "-v")) {
            if (oa.verbosityLevel < 1) oa.verbosityLevel = 1;
        }
        else if (! strcmp(argv[i], "-vv")) {
            if (oa.verbosityLevel < 2) oa.verbosityLevel = 2;
        }
        else if (! strcmp(argv[i], "-h")) {
            std::cerr << "For help, give -h as only parameter.\n";
            return OPTION_ERROR;
        }
        else if (! strcmp(argv[i], "-t")) oa.displayTimeMemory = true;
        else if (! strcmp(argv[i], "-q")) oa.rational = true;
        else if (! strcmp(argv[i], "-r")) {
            oa.reduced = true;
            if ((i < argc - 3) && (oa.redStrand = atoi(argv[i + 1])))
                ++i;
        }
        else if (! strcmp(argv[i], "-g")) oa.optimiseGirth = false;
        else {
            std::cerr << "Unrecognised option \"" << argv[i] << "\".\n";
            return OPTION_ERROR;
        }
    }

    bool goodInput;
    if (! strcmp(argv[argc - 2], "braid"))
        goodInput = braidNotationToVUiv(argv[argc - 1], &(oa.v));
    else if (! strcmp(argv[argc - 2], "dt"))
        goodInput = dtToVUiv(argv[argc - 1], &(oa.v));
    else if (! strcmp(argv[argc - 2], "pd"))
        goodInput = pdToVUiv(argv[argc - 1], &(oa.v));
    else {
        std::cerr << "Next to last parameter has to be \"braid\", \"dt\" or "
            "\"pd\".\n";
        return OPTION_ERROR;
    }
    if (! goodInput) {
        std::cerr << "Error parsing the given diagram.\n";
        return OPTION_ERROR;
    }
    removeAllKinks(oa.v);
    return OPTION_GOOD;
}

bool dtToVUiv(char *&str, VECVECINT *v) {
    VECINT input;
    VECVECINT linkInput;
    if (! strToUiv(str, &input, '-') ) return false;

    linkInput.push_back(input);

    const unsigned nbCrossings = input.size();
    unsigned linkComponents[2 * nbCrossings];
    for (unsigned i = 0; i < 2 * nbCrossings; ++i)
        linkComponents[i] = i;
    {
        int offset = 0;
        for (unsigned i = 0; i < linkInput.size(); ++i) {
            const int newOffset = offset + linkInput.at(i).size();
            linkComponents[2 * newOffset] = 2 * offset;
            offset = newOffset;
        }
    }

    unsigned int c[2 * nbCrossings];
    memset(c, 0, sizeof c);
    unsigned int order[2 * nbCrossings];
    std::vector<bool> input_signs;
    input_signs.resize(nbCrossings);

    for (unsigned i = 0; i < input.size(); ++i) {
        const int j = abs(input.at(i)) - 1;
        if (c[2 * i] || c[j] || (j % 2 == 0) || (j / 2 >= (int)(input.size())))
            return false;
        c[2 * i] = j;
        c[j] = 2 * i;
        input_signs.at(i) = input.at(i) < 0;
        order[2 * i] = i;
        order[j] = i;
    }
    
    std::vector<bool> signs;
    signs.resize(nbCrossings);

    unsigned int region[4 * nbCrossings];
    memset(region, 0, sizeof region);

    UnsignedIntList* position = newUnsignedIntList(0);
    position->next = position;

    if (! dtSigns(position, region, 0, -1, -1, 1, 2, order, input_signs,
                signs, nbCrossings, c, linkComponents, false))
        return false;

    for (unsigned int i = 0; i < input.size(); ++i) {
        VECINT tmp;
        const int inputPlus = abs(input.at(i)) % (input.size() * 2);
        if (input.at(i) > 0) {
            tmp.push_back(abs(input.at(i)) - 1);
            tmp.push_back(signs.at(i) ? 2 * i : (2 * i + 1));
            tmp.push_back(inputPlus);
            tmp.push_back(signs.at(i) ? (2 * i + 1) : 2 * i);
        } else {
            tmp.push_back(2 * i);
            tmp.push_back(signs.at(i) ? (abs(input.at(i)) - 1) : inputPlus);
            tmp.push_back(2 * i + 1);
            tmp.push_back(signs.at(i) ? inputPlus : (abs(input.at(i)) - 1));
        }
        tmp.push_back(signs.at(i) ? NEG_CROS : POS_CROS);
        v->push_back(tmp);
    }
    return true;
}

bool braidNotationToVUiv(char *&str, VECVECINT *v) {
    VECINT first, last;
    unsigned next = 0;
    while (*str) {
        unsigned n;
        bool pos;
        if (('a' <= *str) && (*str <= 'z')) {
            n = *str - 'a';
            pos = true;
        } else if (('A' <= *str) && (*str <= 'Z')) {
            n = *str - 'A';
            pos = false;
        } else return false;
        if (first.size() < n + 2) {
            first.resize(n + 2, -1);
            last.resize(n + 2, -1);
        }
        if (first.at(n) == -1) last.at(n) = first.at(n) = next++;
        if (first.at(n + 1) == -1) last.at(n + 1) = first.at(n + 1) = next++;
        int oldLastN = last.at(n);
        int oldLastNp1 = last.at(n + 1);
        last.at(n) = next++;
        last.at(n + 1) = next++;
        VECINT cros;
        if (pos) {
            cros.push_back(oldLastNp1);
            cros.push_back(last.at(n + 1));
            cros.push_back(last.at(n));
            cros.push_back(oldLastN);
            cros.push_back(POS_CROS);
        } else {
            cros.push_back(oldLastN);
            cros.push_back(oldLastNp1);
            cros.push_back(last.at(n + 1));
            cros.push_back(last.at(n));
            cros.push_back(NEG_CROS);
        }
        v->push_back(cros);
        ++str;
    }

    /* Connect the end of a strand to his beginning */
    for (unsigned i = 0; i < v->size(); ++i)
        for (unsigned j = 0; j < 4; ++j)
            for (unsigned k = 0; k < first.size(); ++k)
                if (v->at(i).at(j) == last.at(k))
                    v->at(i).at(j) = first.at(k);

    /* Decrease numbers so there are no unused numbers */
    int circles = 0;
    for (unsigned k = 0; k < last.size(); ++k)
        if (last.at(k) == -1) ++circles;

    for (unsigned i = 0; i < v->size(); ++i)
        for (unsigned j = 0; j < 4; ++j) {
            int subtract = -circles;
            for (unsigned k = 0; k < last.size(); ++k)
                if (v->at(i).at(j) > last.at(k)) ++subtract;
            v->at(i).at(j) -= subtract;
        }

    next += circles - last.size();

    /* crossingless strands */
    for (unsigned k = 0; k < first.size(); ++k)
        if (first.at(k) == -1) {
            {
                VECINT cros;
                first.at(k) = next++;
                last.at(k) = next++;
                cros.push_back(first.at(k));
                cros.push_back(last.at(k));
                cros.push_back(SEGMENT);
                v->push_back(cros);
            } {
                VECINT cros;
                cros.push_back(last.at(k));
                cros.push_back(first.at(k));
                cros.push_back(SEGMENT);
                v->push_back(cros);
            }
        }
    return true;
}

bool pdToVUiv(char *&str, VECVECINT *v) {
    if (! strToVUiv(str, v)) return false;
    for (VECVECINT::iterator it = v->begin(); it != v->end(); ++it)
        for (VECINT::iterator it2 = it->begin(); it2 != it->end(); ++it2)
            --*it2;
    for (VECVECINT::iterator it = v->begin(); it != v->end(); ++it) {
        int crossingType;
        if (it->at(3) == it->at(1) + 1) crossingType = NEG_CROS;
        else if (it->at(3) == it->at(1) - 1) crossingType = POS_CROS;
        else if (it->at(3) > it->at(1)) crossingType = POS_CROS;
        else crossingType = NEG_CROS;

        /* The following loop resolves the weird PD-anomaly, which causes
         * the above crossing-sign to be potentially wrong if the link is
         * either the hopf link, or the sum of some link with the hopf link.
         */
        for (VECVECINT::iterator it2 = v->begin(); it2 != v->end(); ++it2) {
            if ((crossingType == POS_CROS) && (it2->at(0) == it->at(3))) {
                crossingType = NEG_CROS;
                break;
            }
            if ((crossingType == NEG_CROS) && (it2->at(0) == it->at(1))) {
                crossingType = POS_CROS;
                break;
            }
        }
        it->push_back(crossingType);
    }
    return true;
}

void removeKink(VECVECINT &v, int i, int idx, int &maxStrandNumber) {
    maxStrandNumber += 1;

    VECINT newSeg;
    newSeg.push_back(v.at(i).at(idx));
    newSeg.push_back(maxStrandNumber);
    newSeg.push_back(SEGMENT);
    v.push_back(newSeg);

    int tr[] = {0, 1, 3, 0};
    v.at(i).at(tr[idx]) = maxStrandNumber;
}

void removeAllKinks(VECVECINT &v) {
    int maxStrandNumber = 0;
    for (VECVECINT::iterator it = v.begin(); it != v.end(); ++it)
        for (VECINT::iterator it2 = it->begin(); it2 != it->end(); ++it2)
            if (*it2 > maxStrandNumber) maxStrandNumber = *it2;
    const int vsize = v.size();
    for (int i = 0; i < vsize; ++i)
        if ((v.at(i).back() == POS_CROS) || (v.at(i).back() == NEG_CROS))
            for (int k = 0; k < 4; ++k)
                if (v.at(i).at(k) == v.at(i).at((k + 1) % 4))
                    removeKink(v, i, k, maxStrandNumber);
}


bool strToUiv(char *&str, VECINT *uiv, char prefix) {
    if (!*str) return false;
    if (*(str++) != '[') return false;
    int number = 0;
    bool something = false;
    bool sign = false;
    while ((*str != ']') && (*str != 0)) {
        if ((*str >= '0') && (*str <= '9')) {
            number = number * 10 + *str - '0';
            something = true;
        }
        else if ((*str == ',') && something) {
            uiv->push_back(sign ? (- number) : number);
            sign = false;
            number = 0;
            something = false;
        } else if (prefix && (*str == prefix)) {
            if (sign) return false;
            else sign = true;
        } else return false;
        ++str;
    }
    if (!*str) return false;
    if (something) uiv->push_back(sign ? (- number) : number); 
    ++str;
    return true;
}

bool strToVUiv(char *&str, VECVECINT *v, char prefix) {
    if (!(*str)) return false;
    if (*(str++) != '[') return false;
    while (*str != ']') {
        VECINT newV;
        if (!strToUiv(str, &newV, prefix)) return false;
        v->push_back(newV);
        if (*str == ',') {
            ++str;
            if (*str != '[') return false;
        }
        else if (*str != ']') return false;
    }
    ++str;
    return true;
}
