/*
 *    rat.C --- Wrapper for the arbitrary-precision rationals
 *              (or integers) of the mpir-library.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012/2013 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <mpir.h>

#include "rat.H"

ratio::ratio(const int &other) {
    if (useRationals) {
        mpq_t* frac = (mpq_t*)malloc(sizeof(mpq_t));
        mpq_init(*frac);
        mpq_set_si(*frac, other, 1);
        val = (void*)(frac);
    } else {
        mpz_t* Int = (mpz_t*)malloc(sizeof(mpz_t));
        mpz_init(*Int);
        mpz_set_si(*Int, other);
        val = (void*)(Int);
    }
}

ratio::ratio() {
    if (useRationals) {
        mpq_t* frac = (mpq_t*)malloc(sizeof(mpq_t));
        mpq_init(*frac);
        val = (void*)(frac);
    } else {
        mpz_t* Int = (mpz_t*)malloc(sizeof(mpz_t));
        mpz_init(*Int);
        val = (void*)(Int);
    }
}

ratio::~ratio() {
    if (useRationals) {
        mpq_clear(*((mpq_t*)val));
        free((mpq_t*)val);
    } else {
        mpz_clear(*((mpz_t*)val));
        free((mpz_t*)val);
    }
}

void ratio::operator=(const ratio &other) {
    if (useRationals) 
        mpq_set(*((mpq_t*)val), *((mpq_t*)other.val));
    else
        mpz_set(*((mpz_t*)val), *((mpz_t*)other.val));
}

ratio::ratio(const ratio &other) {
    if (useRationals) {
        mpq_t* frac = (mpq_t*)malloc(sizeof(mpq_t));
        mpq_init(*frac);
        mpq_set(*frac, *((mpq_t*)other.val));
        val = (void*)(frac);
    } else {
        mpz_t* Int = (mpz_t*)malloc(sizeof(mpz_t));
        mpz_init(*Int);
        mpz_set(*Int, *((mpz_t*)other.val));
        val = (void*)(Int);
    }
}

bool ratio::isInvertible() const {
    if (useRationals)
        return isNonZero();
    else 
        return (mpz_cmp_si(*((mpz_t*)val), 1) == 0)
            || (mpz_cmp_si(*((mpz_t*)val), -1) == 0);
}

bool ratio::isNonZero() const {
    if (useRationals)
        return mpq_sgn(*((mpq_t*)val)) != 0;
    else
        return mpz_sgn(*((mpz_t*)val)) != 0;
}

void ratio::inv() {
    if (useRationals)
        mpq_inv(*((mpq_t*)val), *((mpq_t*)val));
}

ratio ratio::operator+=(const ratio& r) {
    if (useRationals)
        mpq_add(*((mpq_t*)val), *((mpq_t*)val), *((mpq_t*)r.val));
    else
        mpz_add(*((mpz_t*)val), *((mpz_t*)val), *((mpz_t*)r.val));
    return *this;
}

ratio ratio::operator*=(const ratio& r) {
    if (useRationals)
        mpq_mul(*((mpq_t*)val), *((mpq_t*)val), *((mpq_t*)r.val));
    else
        mpz_mul(*((mpz_t*)val), *((mpz_t*)val), *((mpz_t*)r.val));
    return *this;
}
