/*
 *    simplifyFoams.C --- Contains the function that simplify Z-linear
 *                        combinations of foams.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012/2013 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <list>
#include <vector>
#include <deque>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"

bool compareBoundaries(const EDGECONT<edge*> &L1, const EDGECONT<edge*> &L2);
bool areBoundariesEqual(const EDGECONT<edge*> &L1, const EDGECONT<edge*> &L2);
bool compareFacets(facet* f1, facet* f2);

void facet::sortOneBoundary(EDGECONT<edge*> &L) {
    if (L.empty()) return;
    EDGECONT<edge*>::iterator min = L.begin();
    for (EDGECONT<edge*>::iterator i = L.begin(); i != L.end(); ++i)
        if ((! (*i)->isInner) && (*i < *min)) {
            bool isTransit[2];
            oEdge *oe = (*i)->asOEdge();
            for (unsigned k = 0; k < 2; ++k)
                isTransit[k] = (! oe->points[k]->isInner) ||
                    ( oe->points[k]->asIPoint()->isTransit());
            if ((!isTransit[0]) && (!isTransit[1])) min = i;
            else if (isTransit[0] || !isTransit[1]) {
                EDGECONT<edge*>::iterator j = i;
                ++j;
                if ((j != L.end()) && (! (*j)->isInner)) {
                    oEdge *nextOe = (*j)->asOEdge();
                    const unsigned k = isTransit[0] ? 0 : 1;
                    if (oe->points[k]->asIPoint()->edges[k] == nextOe) min = i;
                }
            }
        }
    L.splice(L.end(), L, L.begin(), min);
}

bool compareBoundaries(const EDGECONT<edge*> &L1, const EDGECONT<edge*> &L2) {
    if (L2.empty()) return false;
    if (L1.empty()) return true;
    EDGECONT<edge*>::const_iterator i1 = L1.begin(), i2 = L2.begin();
    while (i1 != L1.end() && i2 != L2.end() && ((*i1 == *i2) ||
                ((*i1)->isInner && (*i2)->isInner))) {
        ++i1;
        ++i2;
    }
    if (i2 == L2.end()) return false;
    if (i1 == L1.end()) return true;
    if ((*i1)->isInner && ! (*i2)->isInner) return false;
    if ((*i2)->isInner && ! (*i1)->isInner) return true;
    return (*i1 < *i2);
}  

void facet::sort() {
    for (BOUNDARIESCONT<EDGECONT<edge*> >::iterator lleit = boundaries.begin();
            lleit != boundaries.end(); ++lleit)
        sortOneBoundary(*lleit);
    boundaries.sort(&compareBoundaries);
}

bool areBoundariesEqual(const EDGECONT<edge*> &L1, const EDGECONT<edge*> &L2) {
    if (L1.size() != L2.size()) return false;
    for (EDGECONT<edge*>::const_iterator i1 = L1.begin(), i2 = L2.begin();
            i1 != L1.end(); ++i1, ++i2) {
        if ((*i1)->isInner != (*i2)->isInner) return false;
        if ((! (*i1)->isInner) && (*i1 != *i2)) return false;
    }
    return true;
}

bool facet::isEquivalentTo(const facet& other) const {
    if (boundaries.size() != other.boundaries.size()) return false;
    if (dots != other.dots) return false;
    for (BOUNDARIESCONT<EDGECONT<edge*> >::const_iterator lleit =
            boundaries.begin(), lleit2 = other.boundaries.begin();
            lleit != boundaries.end(); ++lleit, ++lleit2)
        if (! areBoundariesEqual(*lleit, *lleit2)) return false;
    return true;
}

bool compareFacets(facet* f1, facet* f2) {
    if (f2->boundaries.empty()) return false;
    if (f1->boundaries.empty()) return true;
    return compareBoundaries(*f1->boundaries.begin(), *f2->boundaries.begin());
}

void foam::sort() {
    for (FACETCONT<facet*>::iterator i = facets.begin(); i != facets.end(); ++i)
        (*i)->sort();
    facets.sort(&compareFacets);
}

/* Works only if foam is sorted */
bool foam::isMultipleOf(const foam& other) const {
    if (facets.size() != other.facets.size()) return false;
    for (FACETCONT<facet*>::const_iterator lfit = facets.begin(),
            lfit2 = other.facets.begin(); lfit != facets.end(); ++lfit, ++lfit2)
        if (! (*lfit)->isEquivalentTo(**lfit2)) return false;
    return true;
}

bool two::lcfSimplify(MatrixIterator<lcfoams> *it) {
    if ((*it)->simplify()) {
        (*it)->deleteContents();
        delete **it;
        erase(it);
        return true;
    }
    return false;
}

bool lcfoams::simplify() {
    /* 1. Remove foams with coefficient 0
     * 2. Remove foams with facets with more than two points
     * 3. Remove foams which contain spheres with less than two dots, and erase
     *    spheres with two dots (multiplying the coefficient by -1)
     * 4. Remove foams which contain a singular arc around which each facet
     *    has a point, or two facets have two points
     * 5. Erase singular circles by surgery
     */
    FACETCONT<foam*>::iterator f = foams.begin();
    bool eraseIt = false;
    while (f != foams.end()) {
        if (eraseIt) {
            FACETCONT<foam*>::iterator toErase = f++;
            (*toErase)->deleteContents();
            delete *toErase;
            foams.erase(toErase);
            eraseIt = false;
            continue;
        }

        if (! (*f)->coefficient.isNonZero()) {
            eraseIt = true;
            continue;
        }

        FACETCONT<facet*>::iterator i = (*f)->facets.begin();
        while (i != (*f)->facets.end()) {
            if ((*i)->dots > 2) {
                eraseIt = true;
                break;
            }
            if ((*i)->isSphere()) {
                if ((*i)->dots < 2) {
                    eraseIt = true;
                    break;
                } else {
                    (*f)->coefficient *= ratio(-1);
                    FACETCONT<facet*>::iterator facetToErase = i++;
                    delete *facetToErase;
                    (*f)->facets.erase(facetToErase);
                }
            }
            else ++i;
        }
        if (eraseIt) continue;

        for (SINGULARCONT<iEdge*>::iterator s = (*f)->singulars.begin();
                s != (*f)->singulars.end(); ++s) {
            unsigned int dotsOnFacets[3];
            for (unsigned int j = 0; j < 3; ++j)
                dotsOnFacets[j] = (*s)->facets[j]->dots;
            if ((dotsOnFacets[0] + dotsOnFacets[1] + dotsOnFacets[2] > 3) ||
                    (dotsOnFacets[0] && dotsOnFacets[1] && dotsOnFacets[2])) {
                eraseIt = true;
                break;
            }
        }
        if (eraseIt) continue;

        /* 1. Save the three facets
         * 2. Delete the singular arc, and delete it from the facets
         * 3. Make six copies with varying dots distribution
         */
        for (SINGULARCONT<iEdge*>::iterator s = (*f)->singulars.begin();
                s != (*f)->singulars.end(); ++s)
            if ((*s)->isCircle()) {
                facet* facets[3];
                for (int j = 0; j < 3; ++j) {
                    facets[j] = (*s)->facets[j];
                    facets[j]->eraseEdge(**s);
                }
                delete *s;
                (*f)->singulars.erase(s);
                const int addDots[6][3] = { {0, 1, 2}, {2, -1, -1}, {-1, 2, -1},
                    {0, -2, 2}, {-1, 2, -1}, {2, -1, -1} };
                const int signs[6] = { 1, 1, 1, -1, 1, 1 };
                for (int j = 0; j < 6; ++j) {
                    for (int k = 0; k < 3; ++k)
                        facets[k]->dots += addDots[j][k];
                    (*f)->coefficient *= ratio(signs[j]);
                    foams.push_back((*f)->copy());
                }
                eraseIt = true;
                break;
            }
        if (eraseIt) continue;
        ++f;
    }
    if (foams.size() > 1) {
        FACETCONT<foam*>::iterator i = foams.begin();
        for (; i != foams.end(); ++i) (*i)->sort();
        i = foams.begin();
        while (i != foams.end()) {
            FACETCONT<foam*>::iterator j = i;
            ++j;
            while (j != foams.end()) {
                if ((*i)->isMultipleOf(**j)) {
                    (*i)->coefficient += (*j)->coefficient;
                    (*j)->deleteContents();
                    delete *j;
                    j = foams.erase(j);
                } else ++j;
            }
            if (! (*i)->coefficient.isNonZero()) {
                (*i)->deleteContents();
                delete *i;
                i = foams.erase(i);
            }
            else ++i;
        }
    } else if (isInt()) {
        const ratio valueAsInt = sumOfCoefficients();
        for (FACETCONT<foam*>::iterator i = foams.begin();
                i != foams.end(); ++i)
            delete *i;
        foams.clear();
        if (valueAsInt.isNonZero())
            foams.push_back(new foam(valueAsInt));
    }
    return foams.empty();
}
