/*
 *    sparmat.H --- Sparse matrices library. Mainly inspired by Shumakovitch's
 *                  sparmat. All the code is in this file, there is no
 *                  sparmat.C.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012/2013 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

typedef unsigned int length_t;
const length_t minusOne = (length_t)(-1);

template <class E> class SparseEntry {
    public:
        length_t idx;
        E *val;
        SparseEntry(length_t i, E *v) : idx(i), val(v) { }
};

template <class E> class MatrixIterator;

template <class E> class SparseMatrix {
    private:
        typedef ENTRYCONT<SparseEntry<E> > SparseVector;
        typedef typename SparseVector::iterator iter;

        ROWCOLCONT<SparseVector*> rows_;
        ROWCOLCONT<SparseVector*> cols_;

        bool findElement(length_t r, length_t c, iter &rIt, iter &cIt) const;
        bool findElementInVector(SparseVector *v, length_t idx, iter &it) const;
        void resizeDeque(ROWCOLCONT<SparseVector*> &d, length_t l);
        void eraseRowOrCol(length_t idx, ROWCOLCONT<SparseVector*> &toBeErase,
                ROWCOLCONT<SparseVector*> & other, void (E::* kill)());
        length_t undeletedEntries(const ROWCOLCONT<SparseVector*> &v) const;
        void copyVector(length_t r, length_t times,
                ROWCOLCONT<SparseVector*> &thisOne,
                ROWCOLCONT<SparseVector*> & other, E* (E::* cp)() = 0);
    public:
        SparseVector* rowsSM(length_t r) const { return rows_.at(r); }
        SparseVector* colsSM(length_t c) const { return cols_.at(c); }
        length_t rowsSize() const { return rows_.size(); }
        length_t colsSize() const { return cols_.size(); }

        void resize(length_t r, length_t c);
        void addRows(length_t r);
        void addCols(length_t c);
        void copyRows(length_t r, length_t times, E* (E::* cp)());
        void copyCols(length_t c, length_t times, E* (E::* cp)());
        E* at(length_t r, length_t c) const;
        bool erase(length_t r, length_t c, iter *rIt = 0, iter *cIt = 0);
        void erase(MatrixIterator<E> *it, bool keepItReal = false);
        void set(length_t r, length_t c, E* val, bool guaranteedLast = false);
        void eraseRow(length_t r, void (E::* kill)() = 0);
        void eraseCol(length_t c, void (E::* kill)() = 0);
        bool isZero() const;
        void replaceEveryElement(E* (E::* cp)());

        length_t undeletedRows() const { return undeletedEntries(rows_); }
        length_t undeletedCols() const { return undeletedEntries(cols_); }

        MatrixIterator<E> begin(bool keepItReal = false);
};


template <class E> class MatrixIterator {
    friend class SparseMatrix<E>;
    private:
        typedef ENTRYCONT<SparseEntry<E> > SparseVector;
        typedef typename SparseVector::iterator iter;

        ROWCOLCONT<SparseVector*> * rows_;
        ROWCOLCONT<SparseVector*> * cols_;
        iter it;
        length_t idx;
        length_t realR_;
        length_t realC_;

        void setIdxIt(bool keepItReal = false);
    public:
        length_t r() { return idx; }
        length_t c() { return it->idx; }
        length_t realR() { return realR_;}
        length_t realC() { return realC_;}
        E*& operator*()  { return it->val; }
        E* operator->() { return it->val; }

        MatrixIterator() : idx(0), realR_(0), realC_(0) {}

        bool isFinished() { return (! rows_ ); }
        bool step(bool keepItReal = false);
        void setIt(const iter &newIt) { it = newIt; }
};



template <class E> bool MatrixIterator<E>::step(bool keepItReal) {
    length_t oldColIdx = it->idx;
    ++it;
    if (it == rows_->at(idx)->end()) {
        if (keepItReal) ++realR_;
        ++idx; 
        setIdxIt(keepItReal);
        return true;
    } else if (keepItReal) {
        realC_ += it->idx - oldColIdx;
        for (length_t k = oldColIdx + 1; k < it->idx; ++k)
            if (! cols_->at(k)) --realC_;
    }
    return false;
}

template <class E> void MatrixIterator<E>::setIdxIt(bool keepItReal) {
    for (;;) {
        if (idx == rows_->size()) {
            rows_ = 0;
            return;
        }
        if (!rows_->at(idx)) {
            ++idx;
            continue;
        }
        if (rows_->at(idx)->empty()) {
            ++idx;
            ++realR_;
            continue;
        }
        break;
    }
    it = rows_->at(idx)->begin();
    realC_ = it->idx;
    if (keepItReal) 
        for (length_t k = 0; k < it->idx; ++k)
            if (! cols_->at(k)) --realC_;
}

template <class E> void SparseMatrix<E>::eraseRow(
        length_t r, void (E::* kill)()) {
    eraseRowOrCol(r, rows_, cols_, kill);
}

template <class E> void SparseMatrix<E>::eraseCol(
        length_t c, void (E::* kill)()) {
    eraseRowOrCol(c, cols_, rows_, kill);
}

template <class E> MatrixIterator<E> SparseMatrix<E>::begin(bool keepItReal) {
    MatrixIterator<E> result;
    result.rows_ = &rows_;
    result.cols_ = &cols_;
    result.setIdxIt(keepItReal);
    return result;
}

template <class E> length_t SparseMatrix<E>::undeletedEntries(
        const ROWCOLCONT<SparseVector*> &v) const {
    length_t r = 0;
    for (length_t k = 0; k < v.size(); ++k)
        if (v.at(k)) ++r;
    return r;
}

template <class E> bool SparseMatrix<E>::findElementInVector(
        SparseVector *v, length_t idx, iter &it) const {
    if (! v) throw 1;
    it = v->begin();
    while ((it != v->end()) && (it->idx < idx)) ++it;
    return ((it != v->end()) && (it->idx == idx));
}

template <class E> bool SparseMatrix<E>::findElement(length_t r, length_t c,
        iter &rIt, iter &cIt) const {
    bool result1 = findElementInVector(rows_.at(r), c, rIt);
    bool result2 = findElementInVector(cols_.at(c), r, cIt);
    if (result1 != result2) throw 2;
    else {
        if (result1 && (rIt->val != cIt->val)) throw 3; 
        return result1;
    }
}

template <class E> void SparseMatrix<E>::resizeDeque(
        ROWCOLCONT<SparseVector*> &d, length_t l) {
    length_t oldL = d.size();
    for (length_t k = l; k < oldL; ++k) delete d.at(k);
    d.resize(l);
    for (length_t k = oldL; k < l; ++k) d.at(k) = new SparseVector;
}

template <class E> void SparseMatrix<E>::eraseRowOrCol(length_t idx,
        ROWCOLCONT<SparseVector*> &toBeErase,
        ROWCOLCONT<SparseVector*> & other, void (E::* kill)()) {
    if (idx > toBeErase.size()) throw 4;
    SparseVector *theOne = toBeErase.at(idx);
    if (! theOne) throw 5;
    for (iter it1 = theOne->begin(); it1 != theOne->end(); ++it1) {
        iter it2;
        SparseVector *theOther = other.at(it1->idx);
        if (! findElementInVector(theOther, idx, it2)) throw 6;
        if (kill) (it2->val->*kill)();
        theOther->erase(it2);
    }
    delete theOne;
    toBeErase.at(idx) = 0;
}



template <class E> void SparseMatrix<E>::resize(length_t r, length_t c) {
    resizeDeque(rows_, r);
    resizeDeque(cols_, c);
}

template <class E> E* SparseMatrix<E>::at(length_t r, length_t c) const {
    iter rIt, cIt;
    return findElement(r, c, rIt, cIt) ? rIt->val : 0;
}

template <class E> bool SparseMatrix<E>::erase(length_t r, length_t c,
        iter *rIt, iter *cIt) {
    iter rIttmp, cIttmp;
    if (findElement(r, c, rIttmp, cIttmp)) {
        rIttmp = rows_.at(r)->erase(rIttmp);
        cIttmp = cols_.at(c)->erase(cIttmp);
        if (rIt) *rIt = rIttmp;
        if (cIt) *cIt = cIttmp;
        return true;
    } else return false;
}

template <class E> void SparseMatrix<E>::erase(MatrixIterator<E> *it,
        bool keepItReal) {
    const length_t r = it->r();
    const length_t c = it->c();
    const bool iteratorInNextRow = it->step(keepItReal);
    iter rIt;
    if (! erase(r, c, &rIt, 0)) throw 10;
    if (! iteratorInNextRow) it->setIt(rIt);
}

template <class E> void SparseMatrix<E>::replaceEveryElement(E* (E::* cp)()) {
    for (MatrixIterator<E> i = this->begin(); ! i.isFinished(); i.step())
        *i = ((*i)->*cp)();
}

template <class E> void SparseMatrix<E>::set(length_t r, length_t c, E* val,
        bool guaranteedLast) {
    if ((r >= rows_.size()) || (c >= cols_.size())) throw 7;
    if ((! rows_.at(r)) || (! cols_.at(c))) throw 8;
    if (guaranteedLast) {
        rows_.at(r)->push_back(SparseEntry<E>(c, val));
        cols_.at(c)->push_back(SparseEntry<E>(r, val));
    } else {
        iter rIt, cIt;
        if (findElement(r, c, rIt, cIt)) {
            rIt->val = val;
            cIt->val = val;
        } else {
            rows_.at(r)->insert(rIt, SparseEntry<E>(c, val));
            cols_.at(c)->insert(cIt, SparseEntry<E>(r, val));
        }
    }
}

template <class E> bool SparseMatrix<E>::isZero() const {
    for (length_t r = 0; r < rows_.size(); ++r)
        if ( (rows_.at(r)) && (rows_.at(r)->size() > 0) ) return false;
    return true;
}

template <class E> void SparseMatrix<E>::copyVector(length_t idx,
        length_t times, ROWCOLCONT<SparseVector*> &thisOne,
        ROWCOLCONT<SparseVector*> & other, E* (E::* cp)()) {
    if ((idx != minusOne) && (! thisOne.at(idx))) throw 9;
    for (length_t k = 0; k < times; ++k) {
        if (idx == minusOne) thisOne.push_back(new SparseVector);
        else {
            thisOne.push_back(new SparseVector(*thisOne.at(idx)));
            iter j = thisOne.back()->begin();
            while (j != thisOne.back()->end()) {
                j->val = (j->val->*cp)();
                if (j->val) {
                    other.at(j->idx)->push_back(
                            SparseEntry<E>(thisOne.size() - 1, j->val));
                    ++j;
                }
                else j = thisOne.back()->erase(j);
            }
        }
    }
}

template <class E> void SparseMatrix<E>::addRows(length_t r) {
    return copyVector(minusOne, r, rows_, cols_);
}

template <class E> void SparseMatrix<E>::addCols(length_t c) {
    return copyVector(minusOne, c, cols_, rows_);
}

template <class E> void SparseMatrix<E>::copyRows(length_t r,
        length_t times, E* (E::* cp)()) {
    return copyVector(r, times, rows_, cols_, cp);
}

template <class E> void SparseMatrix<E>::copyCols(length_t c,
        length_t times, E* (E::* cp)()) {
    return copyVector(c, times, cols_, rows_, cp);
}
