/*
 *    tensor.C --- Contains the functions for tensoring the chain complex
 *                 with the small chain complex corresponding to a crossing.
 *                      Part of Foamho, a sl3-homology calculator.
 *
 * Copyright (C) 2012/2013 Lukas Lewark <lewark@math.jussieu.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <iostream>
#include <list>
#include <vector>
#include <deque>

#include "macros.H"
#include "rat.H"
#include "sparmat.H"
#include "main.H"

void getPosTensorStuff(web** W1, web** W2, foam** map, oPoint* loIn,
        oPoint* hiOut, oPoint* loOut, oPoint* hiIn);
void getNegTensorStuff(web** W1, web** W2, foam** map, oPoint* loIn,
        oPoint* hiOut, oPoint* loOut, oPoint* hiIn);

void getPosTensorStuff(web** W1, web** W2, foam** map, oPoint* loIn,
        oPoint* hiOut, oPoint* loOut, oPoint* hiIn) {
    *W1 = new web;
    oEdge* e1[2];
    for (int i = 0; i < 2; ++i) {
        e1[i] = new oEdge();
        (*W1)->edges.push_back(e1[i]);
    }
    e1[0]->points[0] = loIn;
    e1[0]->points[1] = hiOut;
    e1[1]->points[0] = hiIn;
    e1[1]->points[1] = loOut;

    *W2 = new web;
    oEdge* e2[5];
    for (int i = 0; i < 5; ++i) {
        e2[i] = new oEdge();
        (*W2)->edges.push_back(e2[i]);
    }
    iPoint* p2[2];
    for (int i = 0; i < 2; ++i) {
        p2[i] = new iPoint();
        (*W2)->points.push_back(p2[i]);
    }
    e2[0]->points[0] = loIn;
    e2[0]->points[1] = p2[0];
    e2[1]->points[0] = hiIn;
    e2[1]->points[1] = p2[0];
    e2[2]->points[0] = p2[1];
    e2[2]->points[1] = p2[0];
    e2[3]->points[0] = p2[1];
    e2[3]->points[1] = loOut;
    e2[4]->points[0] = p2[1];
    e2[4]->points[1] = hiOut;
    p2[0]->edges[0] = e2[0];
    p2[0]->edges[1] = e2[1];
    p2[0]->edges[2] = e2[2];
    p2[1]->edges[0] = e2[4];
    p2[1]->edges[1] = e2[3];
    p2[1]->edges[2] = e2[2];

    *map = new foam;
    (*map)->coefficient = ratio(1);
    iEdge* ie = new iEdge();
    ie->points[0] = p2[1];
    ie->points[1] = p2[0];
    (*map)->singulars.push_back(ie);
    facet* f[3];
    for (int i = 0; i < 3; ++i) {
        f[i] = new facet();
        (*map)->facets.push_back(f[i]);
        ie->facets[i] = f[i];
    }
    EDGECONT<edge*> *l;
    f[0]->boundaries.resize(1);
    l = &*(f[0]->boundaries.begin());
    l->push_back(e1[0]);
    l->push_back(e2[4]);
    l->push_back(ie);
    l->push_back(e2[0]);
    f[1]->boundaries.resize(1);
    l = &*(f[1]->boundaries.begin());
    l->push_back(e2[2]);
    l->push_back(ie);
    f[2]->boundaries.resize(1);
    l = &*(f[2]->boundaries.begin());
    l->push_back(e1[1]);
    l->push_back(e2[3]);
    l->push_back(ie);
    l->push_back(e2[1]);

    (*W1)->qShift = -2;
    (*W2)->qShift = -3;
}

void getNegTensorStuff(web** W1, web** W2, foam** map, oPoint* loIn,
        oPoint* hiOut, oPoint* loOut, oPoint* hiIn) {
    getPosTensorStuff(W1, W2, map, hiIn, loOut, hiOut, loIn);
    web* tmp;
    tmp = *W1;
    *W1 = *W2;
    *W2 = tmp;
    (*W1)->qShift *= -1;
    (*W2)->qShift *= -1;
    (*map)->dualize();
}

void chainComplex::tensorLine(oPoint* in, oPoint* out) {
    web* W = new web();
    oEdge* e = new oEdge();
    W->edges.push_back(e);
    e->points[0] = in;
    e->points[1] = out;
    W->qShift = 0;
    composeEverything(W);
}

void chainComplex::tensorPosCros(oPoint* loIn, oPoint* hiOut, oPoint* loOut,
        oPoint* hiIn) {
    web* W1, *W2;
    foam* map;
    getPosTensorStuff(&W1, &W2, &map, loIn, hiOut, loOut, hiIn);
    tensor(W1, W2, map);

    W1->deleteContents();
    delete W1;
    W2->deleteContents();
    delete W2;
    map->deleteContents();
    delete map;
}

void chainComplex::tensorNegCros(oPoint* loIn, oPoint* hiIn, oPoint* loOut,
        oPoint* hiOut) {
    web* W1, *W2;
    foam* map;
    getNegTensorStuff(&W1, &W2, &map, loIn, hiOut, loOut, hiIn);
    tensor(W1, W2, map);
    --tShift;

    W1->deleteContents();
    delete W1;
    W2->deleteContents();
    delete W2;
    map->deleteContents();
    delete map;
}

void chainComplex::tensor(web* W1, web* W2, foam* map) {
    /* The first part is independent from the parameters */
    maps.push_back(new two);
    maps.back()->resize(0, spaces.back()->webs.size());
    for (unsigned int i = 0; i < maps.back()->colsSize(); ++i)
        if (! spaces.back()->webs.at(i)) maps.back()->eraseCol(i);
    spaces.push_back(new one);
    std::vector<foam*> oldIds;
    std::vector<foam*> ids;
    for (unsigned int spacesI = spaces.size() - 1; spacesI != 0; --spacesI) {
        one* thisOne = spaces.at(spacesI);
        one* nextOne = spaces.at(spacesI - 1);
        two* thisTwo = spacesI < spaces.size() - 1 ? maps.at(spacesI) : 0;
        two* nextTwo = maps.at(spacesI - 1);

        oldIds.swap(ids);
        for (std::vector<foam*>::iterator it = ids.begin();
                it != ids.end(); ++it) {
            if (!*it) continue;
            (*it)->deleteContents();
            delete *it;
        }
        ids.clear();
        ids.reserve(nextOne->realSize());

        const unsigned int oldRowsThis = thisTwo ?
            (thisTwo->rowsSize() - thisOne->realSize()) : 0;

        for (WEBCONT<web*>::iterator nextWebIt = nextOne->webs.begin();
                nextWebIt != nextOne->webs.end(); ++nextWebIt) {
            if (!*nextWebIt) continue;
            ids.push_back(new foam());
            thisOne->webs.push_back((*nextWebIt)->copy(ids.back()));
        }

        const unsigned int oldRowsNext = nextTwo->rowsSize();
        const unsigned int oldColsNext = nextTwo->undeletedCols();
        const unsigned int newRowsNext = nextOne->realSize();

        if (thisTwo) {
            const unsigned int newColsThis = oldColsNext;
            const unsigned int oldColsThis = thisTwo->colsSize();
            thisTwo->addCols(newColsThis);
            for (MatrixIterator<lcfoams> it = nextTwo->begin(true);
                    ! it.isFinished(); it.step(true)) {
                lcfoams* l = it->copy();
                l->glueLcfoams(ids.at(it.realC()), true, false);
                l->glueLcfoams(oldIds.at(it.realR()), false, false);
                l->changeSign();
                thisTwo->set(it.realR() + oldRowsThis, it.realC() + oldColsThis,
                        l, true);
            }
        }

        nextTwo->addRows(newRowsNext);
        for (unsigned int k = 0, colK = 0; k < newRowsNext; ++k, ++colK) {
            while (! nextTwo->colsSM(colK)) ++colK;
            lcfoams* l = new lcfoams(ids.at(k)->copy());
            l->doGlueSS = true;
            nextTwo->set(k + oldRowsNext, colK, l, true);
        }
    }
    for (std::vector<foam*>::iterator it = ids.begin();
            it != ids.end(); ++it) {
        (*it)->deleteContents();
        delete *it;
    }
    for (std::vector<foam*>::iterator it = oldIds.begin();
            it != oldIds.end(); ++it) {
        (*it)->deleteContents();
        delete *it;
    }
    composeEverything(W1, W2, map);
}
